-- WorldPlan
-- WorldQuests.lua
-- Created: 11/2/2016 3:40 PM
-- %file-revision%

local print = DEVIAN_WORKSPACE and function(...) _G.print('WorldQuests', ...) end or function() end
local rprint = DEVIAN_WORKSPACE and function(...) _G.print('WQRefresh', ...) end or function() end
local qprint = DEVIAN_WORKSPACE and function(...) _G.print('POI', ...) end or function() end
local wprint = DEVIAN_WORKSPACE and function(...) _G.print('WP', ...) end or function() end
local mprint = DEVIAN_WORKSPACE and function(...) _G.print('Canvas', ...) end or function() end
local _, db = ...
local Module = WorldPlanQuestsMixin

local _G = _G
local type, tostring, tonumber, pairs, ipairs = type, tostring, tonumber, pairs, ipairs
local MC_GetNumZones, MC_GetZoneInfo = C_MapCanvas.GetNumZones, C_MapCanvas.GetZoneInfo
local TQ_GetQuestsForPlayerByMapID = C_TaskQuest.GetQuestsForPlayerByMapID -- This function is not yet documented
local TQ_GetQuestZoneID = C_TaskQuest.GetQuestZoneID
local TQ_IsActive = C_TaskQuest.IsActive
local pairs, ipairs, tinsert, tremove, wipe = pairs, ipairs, tinsert, tremove, table.wipe
local GetTaskInfo, GetTasksTable, HaveQuestData = GetTaskInfo, GetTasksTable, HaveQuestData
local GetTime = GetTime
local SpellCanTargetQuest, IsQuestIDValidSpellTarget = SpellCanTargetQuest, IsQuestIDValidSpellTarget
local tonumber, abs = tonumber, math.abs
local GetQuestLogRewardInfo = GetQuestLogRewardInfo
local GetCurrentMapAreaID, GetMapInfo, GetMapNameByID = GetCurrentMapAreaID, GetMapInfo, GetMapNameByID
local GetQuestBountyInfoForMapID, GetQuestLogTitle, GetQuestLogIndexByID, IsQuestComplete = GetQuestBountyInfoForMapID, GetQuestLogTitle, GetQuestLogIndexByID, IsQuestComplete
local IsQuestCriteriaForBounty = IsQuestCriteriaForBounty

local ToggleButton = {}
local callbacks = {}
local PinBaseIndex = 1200
local BROKEN_ISLES_ID, DALARAN_ID, AZSUNA_ID, VALSHARAH_ID, HIGHMOUNTAIN_ID, STORMHEIM_ID, SURAMAR_ID, EOA_ID = 1007, 1014, 1015,1018, 1024, 1017, 1033, 1096
local WORLD_QUEST_MAPS = { [DALARAN_ID] = 'Dalaran70',  [AZSUNA_ID] = 'Azsuna',  [VALSHARAH_ID] = "Val'sharah",
  [HIGHMOUNTAIN_ID] = 'Highmountain', [STORMHEIM_ID] = 'Stormheim',  [SURAMAR_ID] = 'Suramar',  [EOA_ID] = 'EyeOfAszhara', }

local REWARD_CASH = WORLD_QUEST_REWARD_TYPE_FLAG_GOLD
local REWARD_ARTIFACT_POWER = WORLD_QUEST_REWARD_TYPE_FLAG_ARTIFACT_POWER
local REWARD_GEAR = WORLD_QUEST_REWARD_TYPE_FLAG_EQUIPMENT
local REWARD_CURRENCY = WORLD_QUEST_REWARD_TYPE_FLAG_ORDER_RESOURCES
local REWARD_REAGENT = WORLD_QUEST_REWARD_TYPE_FLAG_MATERIALS
local SCALE_FACTORS = { 0.25, 0.7, 1 }

local BountyBoard = WorldMapFrame.UIElementsFrame.BountyBoard
local ActionButton = WorldMapFrame.UIElementsFrame.ActionButton

local layoutDirty = true
local bountiesDirty = true
local currentScale = WorldMapDetailFrame:GetScale()
local canTargetQuests
local numShown = 0
local numLoaded = 0
local isDataLoaded = true
local numOverlays = 1
local scaleConstant = 1
Module.TasksByID = {}

--%debug%
local SetTimedCallbackForAllPins = function(seconds, callback)
  C_Timer.After(seconds, function()
    for id, pin in pairs(WorldPlanQuests.QuestsByID) do
      callback(pin)
    end
  end)
end

function Module:OnUpdate(sinceLast)

  if self.filtersDirty or self.isStale then
    self:Refresh()
  end
  if #db.UpdatedPins >= 1 then
    print('|cFF00FF88pending updates', #db.UpdatedPins)
    self:RefreshIfQueued()
  end

end


function Module:Setup()
  print('|cFFFF4400'..self:GetName()..':Setup()')
  for mapID, mapName in pairs(WORLD_QUEST_MAPS) do
    db.QuestsByZone[mapID] = {}
  end
  for target, arg in pairs(callbacks) do
    print(type(target))
    if type(target) == 'table' then
      local callerName = target:GetName() or tostring(target)
      for name, method  in pairs(arg) do
        print(callerName, arg)
        hooksecurefunc(target, name, function(...)
          self:OnSecureHook(callerName .. '.' .. name, method, ...)
        end)
      end
    else
      hooksecurefunc(target, function(...)
        self:OnSecureHook(target, arg, ...)
      end)
    end
  end

  self.Status = CreateFrame('Frame', nil, self)
  self.Status:SetPoint('TOPLEFT', WorldMapPOIFrame, 'TOPLEFT', 0, 0)
  self.Status:SetPoint('BOTTOMRIGHT', WorldMapPOIFrame, 'TOPRIGHT', 0, -4)
  self.Status.t = self.Status:CreateTexture(nil, 'OVERLAY')
  self.Status.b = self.Status:CreateTexture(nil, 'BACKGROUND')
  self.Status.b:SetColorTexture(0,0,0,.25)
  self.Status.b:SetAllPoints(self.Status)
  self.Status.t:SetColorTexture(1,1,1,.5)
  self.Status.t:SetPoint('TOP')
  self.Status.t:SetPoint('BOTTOM')
  self.Status.t:SetPoint('LEFT')
  local translationEnd, translationStart
  self.Status:SetScript('OnUpdate', function(status)
    local translateTo
    if numLoaded < numShown then
      translateTo = numLoaded/numShown * status:GetWidth()
      status.t:SetWidth(translateTo)
    else
      translateTo = numShown * status:GetWidth()
      status.t:SetWidth(translateTo)
    end
  end)

  self:SetAllPoints(WorldMapPOIFrame)
  for k,v in pairs( ToggleButton) do
    self.Toggle:SetScript(k,v)
  end

  self:UpdateBounties('SETUP')

  self:Show()
end
callbacks.ClickWorldMapActionButton = function(WorldQuests)
  WorldQuests:Refresh('CLICK_MAP_ACTION_BUTTON')
end
callbacks.WorldMap_UpdateQuestBonusObjectives = function(WorldQuests)
  WorldQuests:UpdateTaskPOIs()
end
callbacks.WorldMapFrame_UpdateMap = function(WorldQuests)
  WorldQuests:RefreshIfChanged()
end
callbacks.WorldMapScrollFrame_ReanchorQuestPOIs = function (WorldQuests)
  WorldQuests:RefreshIfChanged()
end

callbacks[BountyBoard] = {}
callbacks[BountyBoard].SetSelectedBountyIndex = function(WorldQuests)
  WorldQuests:UpdateBounties('BOUNTY_SELECTED')
  WorldQuests:Refresh('BOUNTY_SELECTED')
end

callbacks[ActionButton] = {}
callbacks[ActionButton].UpdateCastingState = function(WorldQuests)
  WorldQuests:Refresh('CASTING_STATE_CHANGED')
end

function Module:OnConfigUpdate()
  print('|cFFFFFF00OnConfigUpdate()|r')
  if db.Config.FadeWhileGrouped then
    db.PinAlpha = 0.15
  else
    db.PinAlpha = 1
  end

  if not db.Config.EnablePins then
    for _, pin in pairs(db.QuestsByID) do
      pin:SetShown(false)
    end
  end

  ToggleButton.OnShow(self.Toggle)
end

local InternalHideButton = function(button, index)
  button:Hide()
end
local InternalShowButton = function(button, index)
  button:Show()
end

db.UsedPOIs = {}
function Module:UpdateTaskPOIs()
  canTargetQuests = SpellCanTargetQuest()
  local func = canTargetQuests and 'Show' or 'Hide'
  for i = 1, NUM_WORLDMAP_TASK_POIS do
    local button = _G['WorldMapFrameTaskPOI'..i]
    if button and button.worldQuest then
      button[func](button)
    end
  end
end

function Module:OnSecureHook(callbackName, func, ...)
  rprint('|cFFFF4400'..callbackName..'|r', ...)
  func(self, ...)
end

local defaults = {}
local REWARD_UNKNOWN = 768
function Module:OnLoad()
  print('|cFFFF4400'..self:GetName()..':OnLoad()')

  self:SetParent(WorldMapPOIFrame)
  WorldPlan:AddHandler(self, defaults)

  for areaID, fileName in pairs(WORLD_QUEST_MAPS) do
    db.QuestsByZone[areaID] = {}
  end

  -- WORLD_MAP_UPDATE and PLAYER_ENTERING_WORLD are passed down from a higher level
  self:RegisterEvent('WORLD_QUEST_COMPLETED_BY_SPELL')
  self:RegisterEvent('SUPER_TRACKED_QUEST_CHANGED')
  self:RegisterEvent('SKILL_LINES_CHANGED')
  self:RegisterEvent('CURRENT_SPELL_CAST_CHANGED')
  self:RegisterEvent('ARTIFACT_UPDATE')
  self:RegisterEvent('QUEST_LOG_UPDATE')
end

local artifactKnowledgeMultiplier
local superTrackedQuestID
function Module:OnEvent (event, ...)

  print('|cFFFFFF00'..self:GetName()..':OnEvent() '..event..'|r', GetTime(), ...)
  if (event == 'QUEST_LOG_UPDATE') then
    self:UpdateBounties(event)
    self:RefreshIfChanged(event)
  elseif event == 'WORLD_QUEST_COMPLETED_BY_SPELL' then
    local questID = ...
    self:UpdateBounties(event)
    if questID and db.QuestsByID[questID] then
      db.QuestsByID[questID].complete = true
    end
    self:Refresh(event)
  elseif event == 'SKILL_LINES_CHANGED' or event == 'CURRENT_SPELL_CAST_CHANGED' then
    self:Refresh(event)
  elseif event == 'ARTIFACT_UPDATE' then
    local akCheck = C_ArtifactUI.GetArtifactKnowledgeMultiplier()
    if akCheck and (akCheck ~= artifactKnowledgeMultiplier) then
      if artifactKnowledgeMultiplier then
        print('push artifact knowledge update', artifactKnowledgeMultiplier, 'to', akCheck)
        for index, pin in pairs( db.QuestsByID) do
          if pin.rewardType == REWARD_ARTIFACT_POWER then
            db.log(pin.questID .. ' ' .. tostring(pin.title) .. ' Flagged for artifact power.')
            pin.itemNumber = 0
            pin.dataLoaded = nil
          end
        end
      else

        print('artifact knowledge multiplier is known', akCheck)
      end
      artifactKnowledgeMultiplier = akCheck
    end
  elseif event == 'SUPER_TRACKED_QUEST_CHANGED' then
    if superTrackedQuestID and db.QuestsByID[superTrackedQuestID] then
      db.QuestsByID[superTrackedQuestID].isStale = true
    end
    local newID = GetSuperTrackedQuestID()
    if newID and db.QuestsByID[newID] then
      db.QuestsByID[newID].isStale = true
    end
  end
end



local bountyQuests = {}
local bountyInfo = {}
local bountyDisplayLocation, bountyLockedQuestID, selectedBountyIndex, selectedBountyQuestID
function Module:UpdateBounties(...)
  print('|cFF00FF88BountyInfo()|r', ...)
  wipe(db.BountiesByFactionID)
  wipe(bountyQuests)

  db.selectedBounty = nil
  selectedBountyIndex = BountyBoard:GetSelectedBountyIndex()
  bountyInfo, bountyDisplayLocation, bountyLockedQuestID = GetQuestBountyInfoForMapID(db.currentMapID, bountyInfo)
  local numBounties = 0
  for index, info in ipairs(bountyInfo) do
    if info.factionID then
      numBounties = numBounties + 1
      info.index = index
      info.complete =  IsQuestComplete(info.questID)
      if not info.complete then
        db.BountiesByFactionID[info.factionID] = info
        if index == selectedBountyIndex then
          db.selectedBounty = info
          selectedBountyQuestID = info.questID
        end
        print(' ', index, info.factionID, GetQuestLogTitle(GetQuestLogIndexByID(info.questID)), info.complete, (index == selectedBountyIndex) and 'SELECTED' or '')
        bountyQuests[info.questID] = info
      end
    end
  end
  bountiesDirty = nil
end

function Module:OnMapInfo(isBrokenIsle, isZoomedOut, mapAreaID, isNewMap, isMapOpen)
  if isNewMap then
    print('|cFF00FF88OnMapInfo()|r, mapAreaID =', mapAreaID,'visible =', isMapOpen, 'changed =', isNewMap)
    if isMapOpen then
      self:Refresh(true)
    else
      self.isStale = true
    end
  else

    rprint('|cFFFFFF00'..self:GetName()..':OnMapInfo()|r, mapAreaID =', mapAreaID,'visible =', isMapOpen, 'changed =', isNewMap)
  end
end


local totalPins = 0
local TQ_GetQuestLocation = C_TaskQuest.GetQuestLocation
function Module:AcquirePin (info)
  local questID = info.questId
  if not questID then
    return nil
  end

  if not QuestUtils_IsQuestWorldQuest(questID) then
    return nil
  end

  -- if we're grabbing a pin, the filters need to be checked
  local pin = db.QuestsByID[questID]
  if not pin then
    local numFree = #db.FreePins
    if numFree >= 1 then
      pin = tremove(db.FreePins, numFree)
      --print('|cFF00FF00Re-using', pin:GetName())
    else
      totalPins = totalPins + 1
      local name = 'WorldPlanQuestMarker' .. numOverlays
      --print('|cFF00FF00Creating', name)
      pin = CreateFrame('Frame', name, WorldMapPOIFrame, 'WorldPlanQuestPin')

      pin:SetFrameStrata('HIGH')
      pin.GetTypeInfo = function(frame, typeID)
        return self:GetTypeInfo(typeID)
      end
      pin:SetID(totalPins)
      numOverlays = numOverlays + 1
      --pin.iconBorder:SetVertexColor(0,0,0,1)
    end
    pin.questID = questID
    pin.worldQuest = true
    pin.throttle = 1
    pin.isNew = true
    pin.currentWidth = nil
    db.QuestsByID[questID] = pin
    tinsert(db.UsedPins, pin)
  end

  if pin and info then
    pin.inProgress = info.inProgress
    pin.floor = info.floor
    pin.numObjectives = info.numObjectives or 0
    if info.x and info.y then
      pin.x = info.x or pin.x
      pin.y = info.y or pin.y
      --rprint('|cFFFF4400coords|r', info.x, info.y)
    end
  end

  if (not pin.dataLoaded) then
    local dataLoaded = pin:GetData()
    isDataLoaded = (isDataLoaded and dataLoaded)
    WorldPlan.dataFlush = true
  end


  pin.isActive = TQ_IsActive(questID)
  pin:CheckFilterRules()
  rprint(pin:GetID(), pin.filtered, pin.used)

  return pin
end

-- remove from index and add it to the recycling heap
function Module:ReleasePin (pin)

  local id = pin.questID
  if id then
    db.QuestsByID[id] = nil

    for i, zone in pairs(db.QuestsByZone) do
      print('-', i, zone[i])
      zone[id] = nil
    end
    db.TasksByID[id] = nil
  end
  pin.isActive = nil
  pin.dataLoaded = nil
  pin:HideFrames()
  tinsert(db.FreePins, pin)

  WorldPlan.dataFlush = true
  print('|cFF00FF00-'.. (pin.mapID and GetMapNameByID(pin.mapID) or '???') ..'|r', id, pin.title)
end

-- re-anchors and scales pins that have had either of these changed due to data loading delays
function Module:RefreshIfQueued()
  print('|cFF00FF88RefreshIfQueued()')
  local pin = tremove(db.UpdatedPins)
  while pin do
    pin:CheckFilterRules()

    local scaleFactor = SCALE_FACTORS[(pin.dataLoaded and not pin.filtered) and scaleConstant or 1]
    print(pin.title, pin.dataLoaded  and not pin.filtered, scaleFactor)
    if pin.used then
      pin:SetAnchor(nil, pin.x, pin.y, self.hostWidth, self.hostHeight, scaleFactor)
      if pin.isNew then
        pin:OnShow()
      end
    end

    pin = tremove(db.UpdatedPins)

  end
end


local msg = '|cFF00FF88WorldQuests:Refresh()|r|cFF00FFFF'
function Module:Refresh(...)

  rprint(msg, ...)
  if not self:IsVisible() then
    rprint('|cFF00FF88WorldQuests:Refresh()|r', ...)
    self.isStale = true
    return self:MarkAllPins()
  else
    rprint('|cFF00FF88WorldQuests:Refresh()|r', ...)
    print('|cFF00FF88WorldQuests:Refresh()|r', ...)
  end

  if not db.Config.EnablePins then
    numShown = 0
    return
  end
  wprint('  '..msg)

  scaleConstant = db.isContinentMap and 2 or 3
  canTargetQuests = SpellCanTargetQuest()

  for index, pin in pairs(db.QuestsByID) do
    pin.used = nil
  end

  self:UpdateAnchors()
  if bountiesDirty then
    print('  bounties dirty, pushing that')
    self:UpdateBounties()
  end


  -- calculate quests shown
  numShown = 0
  numLoaded = 0
  for questID, pin in pairs(db.QuestsByID) do
    local oV = pin:IsShown()
    if pin.used then
      pin.throttle = 1
      if oV == false then
        rprint('|cFF00FF00cleanup +|r', questID, pin.title)
      end
      pin:SetShown(true)
      numShown = numShown + 1
      if pin.dataLoaded then
        numLoaded = numLoaded + 1
      end

    else
      if oV == true then
        rprint('|cFFFF4400 -|r', questID, pin.title)
      end
      pin.hideReason = "Not used in map area " .. (db.currentMapID)
      pin:HideFrames()
    end
  end



  print(' ', numShown, 'shown,', numLoaded, 'with data')
  if numShown > numLoaded then
    self.Status:Show()
  end


  layoutDirty = nil
  self.isStale = nil
  self.sizesDirty = nil
  self.isZoomDirty = nil
end

local refreshReason
function Module:RefreshIfChanged()
  local scaleCheck = WorldMapDetailFrame:GetScale()
  refreshReason = nil
  if scaleCheck ~= currentScale then
    refreshReason = 'map scale updated'
    self:Refresh('WORLD_MAP_SCALE_CHANGED')
    currentScale = scaleCheck
  elseif self.isStale or layoutDirty then
    refreshReason = 'layout is marked dirty'
  end
  if not refreshReason then
    return
  end


  if self:IsVisible() then
    print('|cFF00FFFFRefreshIfChanged()|r', refreshReason)
    self:Refresh('WORLDMAP_QUEUED')
  else
    rprint('|cFF00FFFFRefreshIfChanged()|r flagging for later refresh', refreshReason)
    self.isStale = true
  end
end

-- update visibility states of all pins
function Module:MarkAllPins(pins)
  print('  |cFFFFFF00'..self:GetName()..':MarkAllPins()|r', pins)
  pins = pins or db.QuestsByID
  for questID, pin in pairs(pins) do
    pin.isStale =  true
    rprint('|cFF00FF00filter', pin.questID, pin.filtered, 'used:', pin.used)
  end
end

function Module:UpdateQuestButton(info, mapID)
  local questID, x, y = info.questId, info.x, info.y
  local pin = self:AcquirePin(info)
  if not pin then
    return
  end


  local doUpdate = pin:GetCriteriaState() or pin.isStale
  --print('  |- ', pin.questID, pin.title)
  rprint('|cFF00FF00update|r', pin.questID, pin.title)

  print(doUpdate)

  if x and y then
    local scaleFactor = SCALE_FACTORS[(pin.dataLoaded and not pin.filtered) and scaleConstant or 1]
    pin:SetFrameLevel(PinBaseIndex+pin:GetID())
    pin.owningFrame = WorldMapFrame
    pin:SetAnchor(WorldMapPOIFrame, x, y, self.hostWidth, self.hostHeight, scaleFactor)
    doUpdate = true
    --tinsert(self.UsedPositions, pin)
  end

  if doUpdate then
    print(' changed', pin.title, pin.isCriteria, pin.isBounty, pin.scaleFactor)
    if pin:IsVisible() then
      --numUpdated = numUpdated + 1
      pin:Refresh('WORLDMAP_REFRESH')
    else
      --numHidden = numHidden + 1
      pin.isStale = true
    end
  end

  if mapID then
    if not db.QuestsByZone[mapID] then
      db.QuestsByZone[mapID] = {}
    end
    db.QuestsByZone[mapID][questID] = pin
  end
end

-- Updates quest markers in taskInfo while associating them with the given map
function Module:UpdateQuestsForMap(taskInfo, mapID)
  rprint('Map', GetMapNameByID(mapID), GetMapNameByID(db.currentMapID))
  for index, info in pairs(taskInfo) do
    self:UpdateQuestButton(info, mapID)
  end
end

-- Used to refresh the visible quest markers
function Module:UpdateAnchors ()
  wipe(self.UsedPositions)
  print('  |cFF00FF00'..self:GetName()..':UpdateAnchors()')
  local hostWidth, hostHeight = WorldMapPOIFrame:GetSize()

  if (hostWidth ~= self.hostWidth) or (hostHeight ~= self.hostHeight) then
    self.hostWidth, self.hostHeight = hostWidth, hostHeight
    layoutDirty = true
  end


  rprint('|cFF00FF00'..self:GetName()..':UpdateAnchors()')
  local mapFileName, textureHeight, textureWidth, isMicroDungeon, microDungeonMapName = GetMapInfo()
  if isMicroDungeon then
    return
  end

  isDataLoaded = true
  local taskInfo = TQ_GetQuestsForPlayerByMapID(db.currentMapID)
  if taskInfo then
    self:UpdateQuestsForMap(taskInfo, db.currentMapID)
  end
  local numZones = MC_GetNumZones(db.currentMapID)
  if numZones then
    for i = 1, numZones do
      local mapAreaID = MC_GetZoneInfo(db.currentMapID, i)
      local taskInfo = TQ_GetQuestsForPlayerByMapID(mapAreaID, db.currentMapID)
      if taskInfo then
        self:UpdateQuestsForMap(taskInfo, mapAreaID)
      end
    end
  end
end

function ToggleButton:OnShow()
  self:SetChecked(db.Config.EnablePins and true or false)
end
function ToggleButton:OnClick()
  print(self:GetChecked())
  db.Config.EnablePins = self:GetChecked()
  _G.WorldPlan:OnConfigUpdate()
end
